from __future__ import absolute_import

try:
    import numpy as _np
    _have_numpy = True
except ImportError:
    _have_numpy = False

from .distob import (scatter, gather, vectorize, apply, call_all, 
                     Remote, proxy_methods, ObjectHub, ObjectEngine, Ref)

engine = None

__version__ = '0.1.5'

if _have_numpy:
    from .arrays import (RemoteArray, DistArray, transpose, rollaxis,
                         expand_dims, concatenate, vstack, hstack, dstack, 
                         split, vsplit, hsplit, dsplit, broadcast_arrays)



"""Example of scattering objects onto different cluster engines.

   (First start an IPython cluster, e.g. by typing 'ipcluster start')
"""
import distob
import numpy as np

arrays = [np.random.randn(10000) for i in xrange(20)]

arrays = distob.scatter(arrays)
#arrays = distob.async_scatter(arrays)
print('after scattering, arrays = %s' % repr(arrays))

# can still use the remote arrays as if they were local 
variances = [a.var() for a in arrays]

#distob.wait(variances)
print('result: %g' % np.sqrt(np.mean(variances)))



"""An empty file where tests should be
"""

import pytest
import distob

#TODO Use pytest fixtures to reliably create an IPython local cluster for tests
#     (check how the IPython test suite does it)

def test_engines():
    pass



# Copyright 2014 Matthew J. Aburn
# 
# This program is free software: you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or 
# (at your option) any later version. See <http://www.gnu.org/licenses/>.

"""
This module implements the core functionality of distob.

Functions:
  scatter(obj)  Distribute an object to remote iPython engines, return a proxy.
  gather(obj)  Fetch back a distributed object, making it local again.
  vectorize(f)  Upgrade normal function f to act in parallel on distributed
    lists or arrays
  call_all(sequence, methodname, *args, **kwargs) Call method on each element. 

Classes:
  Remote   base class, used when creating Remote* proxy classes
  @proxy_methods(base)   class decorator for creating Remote* proxy classes
  ObjectHub  dict interface giving refs to all distributed objects cluster-wide
  ObjectEngine  dict holding the distributed objects of a single IPython engine
  Ref  reference to a (possibly remote) object
"""

from __future__ import absolute_import
from IPython import parallel
import distob
from . import _pylru
import types
import copy
import warnings
import importlib
import collections
import numbers

if distob._have_numpy:
    import numpy as np

# types for compatibility across python 2 and 3
try:
    string_types = basestring
except NameError:
    string_types = str


class Error(Exception):
    pass


class DistobTypeError(Error):
    pass


class DistobValueError(Error):
    pass


class DistobClusterError(Error):
    """Thrown if there is a problem using the cluster that we can't fix"""
    pass


class Ref(object):
    """Reference to a (possibly remote) object.

    a Ref can exist on the same IPython engine that owns the object, or on 
    any python instance that has an ObjectHub (such as the IPython client)

    Attributes:
      engine_id (int): The IPython engine number where the object is held. 
        -1 means that the object is held on the IPython client.
      object_id (str): An object id string that is unique cluster-wide.
      type: type of the remote object
      metadata (tuple, optional): brief extra information about the object 
        may be defined by specific Remote* classes, to help set up access.
    """
    def __init__(self, obj):
        if isinstance(obj, Ref):
            self.engine_id = obj.engine_id
            self.object_id = obj.object_id
            self.type = obj.type
            self.metadata = obj.metadata
            distob.engine.incref(self.object_id)
        else:
            self.engine_id = distob.engine.id
            self.type = type(obj)
            if self.type in distob.engine.proxy_types:
                self.metadata = \
                        distob.engine.proxy_types[self.type].__pmetadata__(obj)
            else:
                self.metadata = None
            self.object_id = '%s:%s@e%d' % (
                self.type.__name__, hex(id(obj)), self.engine_id)
            distob.engine[self.object_id] = obj

    def __getstate__(self):
        # pickled or copied Ref is included in refcount
        distob.engine.incref(self.object_id)
        return (self.engine_id, self.object_id, self.type.__name__,
                self.type.__module__, self.metadata)

    def __setstate__(self, state):
        self.engine_id,self.object_id,typename,modname,self.metadata = state
        module = importlib.import_module(modname)
        self.type = getattr(module, typename)

    def __del__(self):
        distob.engine.decref(self.object_id)

    def __repr__(self):
        return u'<Ref to %s>' % self.object_id


class ObjectEngine(dict):
    """A dict holding the distributed objects of a single IPython engine.

    It maintains cluster-wide reference counts of the objects it holds.
    There is one global instance on each engine.

    For each dict entry,
    Key is an object id string (unique cluster-wide)
    Value is the actual object that is to be accessed remotely.

    Attributes:
      id (int): IPython engine id number of this engine. -1 means this is
        an IPython client, rather than an engine.
      refcounts (dict): Key is an object id string (for an object held on
        this engine). Value is the cluster-wide count of how many Ref
        instances exist pointing to that object.
      proxy_types (dict): When an object's methods are called remotely, some
        return types should be returned by proxy instead of returning a 
        full object. This dict holds mappings {real_type: proxy_type} for 
        these types.
    """
    def __init__(self, engine_id):
        self.id = engine_id 
        self.refcounts = dict()
        self.proxy_types = dict()
        super(ObjectEngine, self).__init__()
        # TODO: The following two lines are a workaround until we can fix issue
        #       #58 in dill re transferring dynamically generated classes.
        for realtype, proxytype in ObjectHub._initial_proxy_types.items():
            self._singleeng_reg_proxy_type(realtype, proxytype)

    # TODO temporary function to be removed when we fix the problems with
    # transfering dynamically generated classes via dill (dill issue #58).
    # In the meantime, use pre-defined classes only.
    def _singleeng_reg_proxy_type(self, real_type, proxy_type):
        self.proxy_types[real_type] = proxy_type

    def incref(self, key):
        self.refcounts[key] += 1

    def decref(self, key):
        if self.refcounts[key] is 1:
            self.__delitem__(key)
        else:
            self.refcounts[key] -= 1
    
    def __setitem__(self, key, value):
        if key in self.keys():
            if value is not self[key]:
                raise DistobValueError('object id not unique: %s' % key)
            self.incref(key)
        else:
            super(ObjectEngine, self).__setitem__(key, value)
            self.refcounts[key] = 1

    def __delitem__(self, key):
        if self.refcounts[key] is 1:
            super(ObjectEngine, self).__delitem__(key)
            del self.refcounts[key]
            print('Cleaned up and deleted object %s' % key)
        else:
            self.decref(key)

    def __repr__(self):
        return '<%s instance on engine %d>:\n%s' % (
            self.__class__.__name__, self.id, repr(self.keys()))

    def _repr_pretty_(self, p, cycle):
        return '<%s instance on engine %d>:\n%s' % (
            self.__class__.__name__, self.id, p.pretty(self.keys()))


# TODO put this back inside ObjectHub after dill issue #58 is resolved.
def _remote_reg_proxy_type(real_t, proxy_t):
    distob.engine.proxy_types[real_t] = proxy_t


class ObjectHub(ObjectEngine):
    """A dict providing references to all distributed objects across the cluster

    Key is an object id string (unique cluster-wide)
    Value is the actual object (if local) or a Ref to the object (if remote)

    Operations supported are get and delete.

    Attributes:
      id (int): IPython engine id number of this engine. -1 means this is
        an IPython client, rather than an engine.
      refcounts (dict): Key is an object id string (for an object held on
        this engine). Value is the cluster-wide count of how many Ref
        instances exist pointing to that object.
      proxy_types (dict): When an object's methods are called remotely, some
        return types should be returned by proxy instead of returning a 
        full object. This dict holds mappings {real_type: proxy_type} for 
        these types.
    """
    _initial_proxy_types = dict()

    def __init__(self, engine_id, client):
        """Make an ObjectHub.
        Args:
          engine_id: -1 if this hub is on the client, otherwise this engine's id
          client: IPython.parallel.client
        """
        self._client = client
        self._dv = client.direct_view(targets='all')
        self._dv.use_dill()
        super(ObjectHub, self).__init__(engine_id)
        # TODO: restore following two lines after issue #58 in dill is fixed.
        #for realtype, proxytype in self.__class__._initial_proxy_types.items():
        #    self._runtime_reg_proxy_type(realtype, proxytype)

    @classmethod
    def register_proxy_type(cls, real_type, proxy_type):
        """Configure engines so that remote methods returning values of type
        `real_type` will instead return by proxy, as type `proxy_type`
        """
        if distob.engine is None:
            cls._initial_proxy_types[real_type] = proxy_type
        elif isinstance(distob.engine, ObjectHub):
            distob.engine._runtime_reg_proxy_type(real_type, proxy_type)
        else:
            # TODO: remove next line after issue #58 in dill is fixed.
            distob.engine._singleeng_reg_proxy_type(real_type, proxy_type)
            pass

    def _runtime_reg_proxy_type(self, real_type, proxy_type):
        #print('about to do runtime reg of %s ' % proxy_type)
        self.proxy_types[real_type] = proxy_type
        ar = self._dv.apply(_remote_reg_proxy_type, real_type, proxy_type)
        self._dv.wait(ar)
        pass

    def incref(self, key):
        engine_id = int(key[(key.rindex('@e')+2):])
        if engine_id is self.id:
            super(ObjectHub, self).incref(key)
        else:
            self._dv.execute('distob.engine.incref("%s")' % key, 
                             targets=engine_id)

    def decref(self, key):
        engine_id = int(key[(key.rindex('@e')+2):])
        if engine_id is self.id:
            super(ObjectHub, self).decref(key)
        else:
            self._dv.execute('distob.engine.decref("%s")' % key, 
                             targets=engine_id)

    def __delitem__(self, key):
        engine_id = int(key[(key.rindex('@e')+2):])
        if engine_id is self.id:
            super(ObjectHub, self).__delitem__(key)
        else:
            self._dv.execute('del distob.engine["%s"]' % key, targets=engine_id)

    def __getitem__(self, key):
        engine_id = int(key[(key.rindex('@e')+2):])
        if engine_id is self.id:
            return super(ObjectHub, self).__getitem__(key)
        else:
            def fetch_object(object_id):
                return Ref(distob.engine[object_id])
            self._dv.targets = engine_id
            res = self._dv.apply_sync(fetch_object, key)
            self._dv.targets = 'all'
            return res

    def __setitem__(self, key, value):
        engine_id = int(key[(key.rindex('@e')+2):])
        if engine_id is self.id:
            super(ObjectHub, self).__setitem__(key, value)
        else:
            raise Error('Setting remote objects via ObjectHub not implemented')

    def keys(self):
        local_keys = super(ObjectHub, self).keys()
        def fetch_keys():
            return distob.engine.keys()
        ar = self._dv.apply_async(fetch_keys)
        self._dv.wait(ar)
        remote_keys = [key for sublist in ar.r for key in sublist]
        return remote_keys + local_keys


def _remote_setup_engine(engine_id):
    if distob.engine is None:
        distob.engine = distob.ObjectEngine(engine_id)
    # TODO these imports should be unnecessary with improved deserialization
    import numpy as np
    from scipy import stats
    # TODO Using @parallel.interactive still did not import to __main__
    #      so will do it this way for now.
    import __main__
    __main__.__dict__['np'] = np
    __main__.__dict__['stats'] = stats


def _setup_engines(client=None):
    """Prepare all iPython engines for distributed object processing.

    Args:
      client (IPython.parallel.client, optional): If None, will create a client         using the default IPython profile.
    """
    if not client:
        try:
            client = parallel.Client()
        except:
            raise DistobClusterError(
                u"""Could not connect to an IPython parallel cluster. Make
                 sure a cluster is started (e.g. to use the CPUs of a
                 single computer, can type 'ipcluster start')""")
    ids = client.ids
    if not ids:
        raise DistobClusterError(u'No IPython compute engines are available')
    dv = client[ids]
    dv.use_dill()
    with dv.sync_imports(quiet=True):
        import distob
    # create global ObjectEngine distob.engine on each engine
    ars = []
    for i in ids:
        dv.targets = i
        ars.append(dv.apply_async(_remote_setup_engine, i))
    dv.wait(ars)
    for ar in ars:
        if not ar.successful():
            raise ar.r
    # create global ObjectHub distob.engine on the client host
    if distob.engine is None:
        distob.engine = ObjectHub(-1, client)


class Remote(object):
    """Base class for Remote* proxy classes.
    An instance of a Remote subclass is a proxy designed to control
    an instance of an existing class which may be local or on a remote host.

    To make a Remote proxy class, inherit from `Remote` as first base class.
    If the instances of the new Remote* proxy class should be treated 
    exactly like instances of the controlled class, then also inherit from 
    the controlled class as the second base:

        @proxy_methods(Tree)
        class RemoteTree(Remote, Tree)

    Use the decorator @proxy_methods() to register the new Remote proxy class
    and to specify which methods/attributes of the existing class should be
    proxied. (by default all except those starting with an underscore)

    When each instance is created, it will also check for any instance-specific
    methods and attributes that exist on the controlled object.
    (For example attributes created by the __init__ of the controlled object.)
    Corresponding methods/attributes will be added to the Remote* instance to
    provide remote access to these as well.
    """
    _include_underscore = ()
    _exclude = ()

    def __init__(self, obj):
        """Set up the Remote* proxy object to access an already-existing object,
        which may be local or remote.

        Args:
          obj (Ref or object): either a Ref reference to the (possibly remote) 
            object to be controlled, or else an actual (local) object to be 
            controlled.
        """
        if distob.engine is None:
            _setup_engines()
        if isinstance(obj, Ref):
            self._ref = obj
            self.is_local = (self._ref.engine_id is distob.engine.id)
        else:
            self._ref = Ref(obj)
            self.is_local = True
        if self.is_local:
            self._dv = None
            self._obcache = distob.engine[self._ref.object_id]
            self._obcache_current = True
        else:
            self._dv = distob.engine._client[self._ref.engine_id]
            self._dv.use_dill()
            self._obcache = None
            self._obcache_current = False
        self._id = self._ref.object_id
        #Add proxy controllers for any instance-specific methods/attributes:
        (instance_methods, instance_attribs) = self._scan_instance()
        for name, doc in instance_methods:
            def make_proxy_method(method_name, doc):
                def method(self, *args, **kwargs):
                    if self._obcache_current:
                        return apply(getattr(self._obcache, method_name),
                                     *args, **kwargs)
                    else:
                        return cls._try_cached_apply(self, method_name,
                                                     *args, **kwargs)
                method.__doc__ = doc
                method.__name__ = method_name
                #return types.MethodType(method, None, cls)
                return method
            setattr(self, name, make_proxy_method(name, doc))
        for name in instance_attribs:
            def make_property(attrib_name):
                # TODO: implement fset and fdel (requires writeback cache)
                def getter(self):
                    if self._obcache_current:
                        return getattr(self._obcache, attrib_name)
                    else:
                        return self._cached_apply('__getattribute__', 
                                                  attrib_name) 
                prop = property(fget=getter)
                return prop
            setattr(self.__class__, name, make_property(name))

    @classmethod
    def _local_scan_instance(cls, object_id, include_underscore, exclude):
        method_info = []
        attributes = []
        obj = distob.engine[object_id]
        if hasattr(obj, '__dict__'):
            for name in obj.__dict__:
                if (name not in exclude and 
                    (name[0] != '_' or 
                     include_underscore is True or
                     name in include_underscore)):
                    f = obj.__dict__[name]
                    if callable(f) and not isinstance(f, type):
                        method_info.append((name, f.__doc__))
                    else:
                        attributes.append(name)
        return (method_info, attributes)

    def _scan_instance(self):
        """get information on instance-methods/attributes of the object"""
        if not self.is_local:
            return self._dv.apply_sync(Remote._local_scan_instance,
                                       self._ref.object_id,
                                       self.__class__._include_underscore,
                                       self.__class__._exclude)

    def _apply(self, method_name, *args, **kwargs):
        """Call a method on the remote object without caching."""
        #print('in _apply: type=%s, method_name==%s, args==%s, kwargs==%s' % (
        #        type(self), method_name, args, kwargs))
        def remote_call_method(object_id, method_name, *args, **kwargs):
            obj = distob.engine[object_id]
            result = obj.__getattribute__(method_name)(*args, **kwargs)
            if type(result) in distob.engine.proxy_types:
                return Ref(result)
            else:
                return result
        r = self._dv.apply_sync(remote_call_method, self._id, 
                                method_name, *args, **kwargs)
        if isinstance(r, Ref):
            RemoteClass = distob.engine.proxy_types[r.type]
            return RemoteClass(r)
        else:
            return r

    @_pylru.lrudecorator(1000)
    def _cached_apply(self, method_name, *args, **kwargs):
        #print('cache miss. self:%s(%s); method:%s; args:%s; kwargs:%s' % (
        #    hex(id(self)), self._ref.object_id, method_name, 
        #    repr(args), repr(kwargs)))
        return self._apply(method_name, *args, **kwargs)

    def _try_cached_apply(self, method_name, *args, **kwargs):
        """Call a method on the remote object. Cache results if args hashable.
        Args:
          method_name (str): name of the remote method to call 
          *args, **kwargs: arguments to pass to the remote method
        Returns:
          ``<remote_object>.<method_name>(*args, **kwargs)``
          If the output is of a type that should be proxied, returns 
          a ``Remote*`` proxy object instead of the real object.
        """
        #print('try cached: self:%s(%s); method:%s; args:%s; kwargs:%s' % (
        #   hex(id(self)), self._ref.object_id, method_name, 
        #   repr(args), repr(kwargs)))
        #
        #For some reason immutable slice objects are not hashable in python.
        #TODO workaround this by recursively replacing any slices in dict key.
        try:
            return self._cached_apply(method_name, *args, **kwargs)
        except TypeError as te:
            if te.args[0][:15] == 'unhashable type':
                #print("unhashable. won't be able to cache")
                return self._apply(method_name, *args, **kwargs)
            else:
                raise

    def _apply_async(self, method_name, *args, **kwargs):
        """Call a method on the remote object without caching.
        Returns: 
          ipython.parallel.AsyncResult: async output of the remote method
        """
        def remote_call_method(object_id, method_name, *args, **kwargs):
            obj = distob.engine[object_id]
            result = obj.__getattribute__(method_name)(*args, **kwargs)
            if type(result) in distob.engine.proxy_types:
                return Ref(result)
            else:
                return result
        ar = self._dv.apply_async(remote_call_method, self._id, 
                                  method_name, *args, **kwargs)
        return ar
        if isinstance(r, Ref):
            RemoteClass = distob.engine.proxy_types[r.type]
            return RemoteClass(r)
        else:
            return r

    def _fetch(self):
        """update local cached copy of the real object"""
        if not self.is_local and not self._obcache_current:
            #print('fetching data from %s' % self._ref.object_id)
            self._obcache = self._dv['distob.engine["%s"]' % self._id]
            self._obcache_current = True

    def __ob(self):
        """return a local copy of the real object"""
        self._fetch()
        return self._obcache

    _ob = property(fget=__ob)

    def __distob_gather__(self):
        return self._ob

    def __copy__(self):
        newref = copy.copy(self._ref)
        obj = self.__class__.__new__(self.__class__, newref)
        obj.__init__(newref)
        return obj

    def __deepcopy__(self, memo):
        return self.__copy__()

    @classmethod
    def __pmetadata__(cls, obj):
        """Subclasses can override this method if they need Ref to provide
        more metadata on the controlled object to help configure remote access.
        Arguments:
          obj: object of type ``self._ref.type``
        """
        return None


def proxy_methods(base, include_underscore=None, exclude=None, supers=True):
    """class decorator. Modifies `Remote` subclasses to add proxy methods and
    attributes that mimic those defined in class `base`.

    Example:

        @proxy_methods(Tree)
        class RemoteTree(Remote, Tree)

    The decorator registers the new proxy class and specifies which methods
    and attributes of class `base` should be proxied via a remote call to
    a real object, and which methods/attributes should not be proxied but
    instead called directly on the instance of the proxy class.

    By default all methods and attributes of the class `base` will be
    proxied except those starting with an underscore.

    The MRO of the decorated class is respected:
    Any methods and attributes defined in the decorated class
    (or in other bases of the decorated class that do not come after `base`
     in its MRO) will override those added by this decorator,
    so that `base` is treated like a base class.

    Args:
      base (type): The class whose instances should be remotely controlled.
      include_underscore (sequence of str): Names of any methods or attributes 
        that start with an underscore but should be proxied anyway.
      exclude (sequence of str): Names of any methods or attributes that 
        should not be proxied.
      supers (bool): Proxy methods and attributes defined in superclasses 
        of ``base``, in addition to those defined directly in class ``base``
    """
    if isinstance(include_underscore, str):
        include_underscore = (include_underscore,)
    if isinstance(exclude, str):
        exclude = (exclude,)
    if not include_underscore:
        include_underscore = ()
    if not exclude:
        exclude = ()
    def rebuild_class(cls):
        # Identify any bases of cls that do not come after `base` in the list:
        bases_other = list(cls.__bases__)
        if bases_other[-1] is object:
            bases_other.pop()
        if base in bases_other:
            bases_other = bases_other[:bases_other.index(base)]
        if not issubclass(cls.__bases__[0], Remote):
            raise DistobTypeError('First base class must be subclass of Remote')
        if not issubclass(base, object):
            raise DistobTypeError('Only new-style classes currently supported')
        dct = cls.__dict__.copy()
        if cls.__doc__ is None or '\n' not in cls.__doc__:
            base_doc = base.__doc__
            if base_doc is None:
                base_doc = ''
            dct['__doc__'] = """Local object representing a remote %s
                  It can be used just like a %s object, but behind the scenes 
                  all requests are passed to a real %s object on a remote host.

                  """ % ((base.__name__,)*3) + base_doc
        newcls = type(cls.__name__, cls.__bases__, dct)
        newcls._include_underscore = include_underscore
        newcls._exclude = exclude
        if supers:
            proxied_classes = base.__mro__[:-1]
        else:
            proxied_classes = (base,)
        for c in proxied_classes:
            for name in c.__dict__:
                def mk_proxy_method(method_name, doc):
                    def method(self, *args, **kwargs):
                        if self._obcache_current:
                            return getattr(self._obcache, method_name)(
                                    *args, **kwargs)
                        else:
                            return newcls._try_cached_apply(self, method_name,
                                                            *args, **kwargs)
                    method.__doc__ = doc
                    method.__name__ = method_name
                    #return types.MethodType(method, None, newcls)
                    return method
                def mk_property(attrib_name):
                    # TODO: implement fset and fdel (requires writeback cache)
                    def getter(self):
                        if self._obcache_current:
                            return getattr(self._obcache, attrib_name)
                        else:
                            return self._cached_apply('__getattribute__',
                                                      attrib_name)
                    prop = property(fget=getter)
                    return prop
                #respect MRO: proxy an attribute only if it is not overridden
                if (name not in newcls.__dict__ and
                        all(name not in b.__dict__ 
                            for c in bases_other for b in c.mro()[:-1]) and
                        name not in newcls._exclude and
                        (name[0] != '_' or 
                         newcls._include_underscore is True or
                         name in newcls._include_underscore)):
                    f = c.__dict__[name]
                    if callable(f) and not isinstance(f, type):
                        setattr(newcls, name, mk_proxy_method(name, f.__doc__))
                    else:
                        setattr(newcls, name, mk_property(name))
        #newcls.__module__ = '__main__' # cause dill to pickle it whole
        #import __main__
        #__main__.__dict__[newcls.__name__] = newcls # for dill..
        ObjectHub.register_proxy_type(base, newcls)
        return newcls
    return rebuild_class


def _async_scatter(obj):
    """Distribute an obj or list to remote engines. 
    Return an async result or (possibly nested) lists of async results, 
    each of which is a Ref
    """
    #TODO Instead of special cases for strings and Remote, should have a
    #     list of types that should not be proxied, inc. strings and Remote.
    if (isinstance(obj, Remote) or 
            isinstance(obj, numbers.Number) or 
            obj is None):
        return obj
    if (isinstance(obj, collections.Sequence) and 
            not isinstance(obj, string_types)):
        ars = []
        for i in range(len(obj)):
            ars.append(_async_scatter(obj[i]))
        return ars
    else:
        if distob.engine is None:
            _setup_engines()
        client = distob.engine._client
        dv = distob.engine._dv
        def remote_put(obj):
            return Ref(obj)
        dv.targets = _async_scatter.next_engine
        ar_ref = dv.apply_async(remote_put, obj)
        _async_scatter.next_engine = (
                _async_scatter.next_engine + 1) % len(client)
        dv.targets = client.ids
        return ar_ref

_async_scatter.next_engine = 0


def _ars_to_proxies(ars):
    """wait for async results and return proxy objects
    Args: 
      ars: AsyncResult (or sequence of AsyncResults), each result type ``Ref``.
    Returns:
      Remote* proxy object (or list of them)
    """
    if isinstance(ars, Remote) or ars is None:
        return ars
    elif isinstance(ars, collections.Sequence):
        res = []
        for i in range(len(ars)):
            res.append(_ars_to_proxies(ars[i]))
        return res
    elif isinstance(ars, parallel.AsyncResult):
        ars.wait()
        ref = ars.r
        ObClass = ref.type
        if ObClass in distob.engine.proxy_types:
            RemoteClass = distob.engine.proxy_types[ObClass]
        else:
            RemoteClass = type(
                    'Remote' + ObClass.__name__, (Remote, ObClass), dict())
            RemoteClass = proxy_methods(ObClass)(RemoteClass)
        proxy_obj = RemoteClass(ref)
        return proxy_obj
    else:
        raise DistobTypeError('Unpacking ars: unexpected type %s' % type(ars))


def _scatter_ndarray(ar, axis=-1):
    """Turn a numpy ndarray into a distributed array"""
    from .arrays import DistArray, RemoteArray
    shape = ar.shape
    ndim = len(shape)
    if axis is None:
        return scatter([ar])[0]
    if axis < -ndim or axis > ndim - 1:
        raise DistobValueError('axis out of range')
    if axis < 0:
        axis = ndim + axis
    n = shape[axis]
    if n == 1:
        return scatter([ar])[0]
    if distob.engine is None:
        _setup_engines()
    num_engines = len(distob.engine._client)
    if n > num_engines * 100:
        message = (u'Currently have only implemented splitting an axis into '
                    'chunks of length 1. axis %d is long (%d), so may be slow '
                    'distributing the array along this axis.' % (axis, n))
        warnings.warn(message, RuntimeWarning)
    if isinstance(ar, DistArray):
        if axis == ar._distaxis:
            return ar
        else:
            raise DistobError('Currently can only scatter one axis of array')
    # Currently, if requested to scatter an array that is already Remote, will
    # first get the whole array locally, then scatter. Not really optimal.
    if isinstance(ar, RemoteArray):
        ar = ar._ob
    s = slice(None)
    subarrays = []
    for i in range(n):
        index = (s,)*axis + (slice(i,i+1),) + (s,)*(ndim - axis - 1)
        subarrays.append(ar[index])
    subarrays = scatter(subarrays)
    return DistArray(subarrays, axis)


def scatter(obj, axis=-1):
    """Distribute obj or list to remote engines, returning proxy objects
    Args:
      obj: any python object, or list of objects
      axis (int, optional): currently only used if scattering a numpy array,
        specifying along which axis to split the array to distribute it. The 
        default is to split along the last axis.
    """
    if hasattr(obj, '__distob_scatter__'):
        return obj.__distob_scatter__(axis)
    if distob._have_numpy and (isinstance(obj, np.ndarray) or
                        hasattr(type(obj), '__array_interface__')):
        return _scatter_ndarray(obj, axis)
    if axis is not -1:
        raise DistobValueError('`axis` argument only applies to ndarrays')
    elif isinstance(obj, Remote):
        return obj
    ars = _async_scatter(obj)
    proxy_obj = _ars_to_proxies(ars)
    return proxy_obj


def gather(obj):
    """Retrieve objects that have been distributed, making them local again"""
    if hasattr(obj, '__distob_gather__'):
        return obj.__distob_gather__()
    elif (isinstance(obj, collections.Sequence) and 
            not isinstance(obj, string_types)):
        return [gather(subobj) for subobj in obj]
    else:
        return obj


def vectorize(f):
    """Upgrade normal function f to act in parallel on distibuted lists/arrays

    Args:
      f (callable): an ordinary function which expects as its first argument a
        single object, or a numpy array of N dimensions.

    Returns:
      vf (callable): new function that takes as its first argument a list of
        objects, or a array of N+1 dimensions. ``vf()`` will do the
        computation ``f()`` on each part of the input in parallel and will
        return a list of results, or a distributed array of results.
    """
    def vf(obj, *args, **kwargs):
        # user classes can customize how to vectorize a function:
        if hasattr(obj, '__distob_vectorize__'):
            return obj.__distob_vectorize__(f)(obj, *args, **kwargs)
        def remote_f(object_id, *args, **kwargs):
            result = f(distob.engine[object_id], *args, **kwargs)
            if type(result) in distob.engine.proxy_types:
                return Ref(result)
            else:
                return result
        if isinstance(obj, Remote):
            dv = distob.engine._client[obj._ref.engine_id]
            r = dv.apply_sync(remote_f, obj._ref.object_id, *args, **kwargs)
            if isinstance(r, Ref):
                RemoteClass = distob.engine.proxy_types[r.type]
                return RemoteClass(r)
            else:
                return r
        elif distob._have_numpy and (isinstance(obj, np.ndarray) or
                 hasattr(type(obj), '__array_interface__')):
            distarray = scatter(obj, axis=None)
            return vf(distarray, *args, **kwargs)
        elif isinstance(obj, collections.Sequence):
            inputs = scatter(obj)
            dv = distob.engine._client[:]
            results = []
            for obj in inputs:
                if isinstance(obj, Remote):
                    dv.targets = obj._ref.engine_id
                    results.append(dv.apply_async(remote_f, obj._ref.object_id,
                                                  *args, **kwargs))
                else:
                    results.append(f(obj, *args, **kwargs))
            for i in range(len(results)):
                ar = results[i]
                if isinstance(ar, parallel.AsyncResult):
                    ar.wait()
                    results[i] = ar.r
                if isinstance(results[i], Ref):
                    ref = results[i]
                    RemoteClass = distob.engine.proxy_types[ref.type]
                    results[i] = RemoteClass(ref)
            return results
    if hasattr(f, '__name__'):
        vf.__name__ = 'v' + f.__name__
        f_str = f.__name__ + '()'
    else:
        f_str = 'callable'
    doc = u"""Apply %s in parallel to a list or array\n
           Args:
             obj (Sequence of objects or an array)
             other args are the same as for %s
           """ % (f_str, f_str)
    if hasattr(f, '__doc__') and f.__doc__ is not None:
        doc = doc.rstrip() + (' detailed below:\n----------\n' + f.__doc__)
    vf.__doc__ = doc
    return vf


def apply(f, obj, *args, **kwargs):
    """Apply a function in parallel to each element of the input"""
    return vectorize(f)(obj, *args, **kwargs)


def call_all(sequence, method_name, *args, **kwargs):
    """Call a method on each element of the sequence, in parallel.
    Returns:
      list of results
    """
    # dispatch method calls asynchronously
    results = []
    for obj in sequence:
        if isinstance(obj, Remote):
            results.append(obj._apply_async(method_name, *args, **kwargs))
        else: 
            results.append(getattr(obj, method_name)(*args, **kwargs))
    # now wait for all results before returning
    for i in range(len(results)):
        obj = results[i]
        if isinstance(obj, parallel.AsyncResult):
            obj.wait()
            results[i] = obj.r
        if isinstance(results[i], Ref):
            ref = results[i]
            RemoteClass = distob.engine.proxy_types[ref.type]
            results[i] = RemoteClass(ref)
    return results




# Cache implementaion with a Least Recently Used (LRU) replacement policy and
# a basic dictionary interface.

# Copyright (C) 2006, 2009, 2010, 2011 Jay Hutchinson

# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.



# The cache is implemented using a combination of a python dictionary (hash
# table) and a circular doubly linked list. Items in the cache are stored in
# nodes. These nodes make up the linked list. The list is used to efficiently
# maintain the order that the items have been used in. The front or head of
# the list contains the most recently used item, the tail of the list
# contains the least recently used item. When an item is used it can easily
# (in a constant amount of time) be moved to the front of the list, thus
# updating its position in the ordering. These nodes are also placed in the
# hash table under their associated key. The hash table allows efficient
# lookup of values by key.

# Class for the node objects.
class _dlnode(object):
    def __init__(self):
        self.empty = True


class lrucache(object):

    def __init__(self, size, callback=None):

        self.callback = callback

        # Create an empty hash table.
        self.table = {}

        # Initialize the doubly linked list with one empty node. This is an
        # invariant. The cache size must always be greater than zero. Each
        # node has a 'prev' and 'next' variable to hold the node that comes
        # before it and after it respectivly. Initially the two variables
        # each point to the head node itself, creating a circular doubly
        # linked list of size one. Then the size() method is used to adjust
        # the list to the desired size.

        self.head = _dlnode()
        self.head.next = self.head
        self.head.prev = self.head

        self.listSize = 1

        # Adjust the size
        self.size(size)


    def __len__(self):
        return len(self.table)

    def clear(self):
        for node in self.dli():
            node.empty = True
            node.key = None
            node.value = None

        self.table.clear()


    def __contains__(self, key):
        return key in self.table

    # Looks up a value in the cache without affecting cache order.
    def peek(self, key):
        # Look up the node
        node = self.table[key]
        return node.value


    def __getitem__(self, key):
        # Look up the node
        node = self.table[key]

        # Update the list ordering. Move this node so that is directly
        # proceeds the head node. Then set the 'head' variable to it. This
        # makes it the new head of the list.
        self.mtf(node)
        self.head = node

        # Return the value.
        return node.value

    def get(self, key, default=None):
        """Get an item - return default (None) if not present"""
        try:
            return self[key]
        except KeyError:
            return default

    def __setitem__(self, key, value):
        # First, see if any value is stored under 'key' in the cache already.
        # If so we are going to replace that value with the new one.
        if key in self.table:

            # Lookup the node
            node = self.table[key]

            # Replace the value.
            node.value = value

            # Update the list ordering.
            self.mtf(node)
            self.head = node

            return

        # Ok, no value is currently stored under 'key' in the cache. We need
        # to choose a node to place the new item in. There are two cases. If
        # the cache is full some item will have to be pushed out of the
        # cache. We want to choose the node with the least recently used
        # item. This is the node at the tail of the list. If the cache is not
        # full we want to choose a node that is empty. Because of the way the
        # list is managed, the empty nodes are always together at the tail
        # end of the list. Thus, in either case, by chooseing the node at the
        # tail of the list our conditions are satisfied.

        # Since the list is circular, the tail node directly preceeds the
        # 'head' node.
        node = self.head.prev

        # If the node already contains something we need to remove the old
        # key from the dictionary.
        if not node.empty:
            if self.callback is not None:
                self.callback(node.key, node.value)
            del self.table[node.key]

        # Place the new key and value in the node
        node.empty = False
        node.key = key
        node.value = value

        # Add the node to the dictionary under the new key.
        self.table[key] = node

        # We need to move the node to the head of the list. The node is the
        # tail node, so it directly preceeds the head node due to the list
        # being circular. Therefore, the ordering is already correct, we just
        # need to adjust the 'head' variable.
        self.head = node


    def __delitem__(self, key):

        # Lookup the node, then remove it from the hash table.
        node = self.table[key]
        del self.table[key]

        node.empty = True

        # Not strictly necessary.
        node.key = None
        node.value = None

        # Because this node is now empty we want to reuse it before any
        # non-empty node. To do that we want to move it to the tail of the
        # list. We move it so that it directly preceeds the 'head' node. This
        # makes it the tail node. The 'head' is then adjusted. This
        # adjustment ensures correctness even for the case where the 'node'
        # is the 'head' node.
        self.mtf(node)
        self.head = node.next

    def __iter__(self):

        # Return an iterator that returns the keys in the cache in order from
        # the most recently to least recently used. Does not modify the cache
        # order.
        for node in self.dli():
            yield node.key

    def items(self):

        # Return an iterator that returns the (key, value) pairs in the cache
        # in order from the most recently to least recently used. Does not
        # modify the cache order.
        for node in self.dli():
            yield (node.key, node.value)

    def keys(self):

        # Return an iterator that returns the keys in the cache in order from
        # the most recently to least recently used. Does not modify the cache
        # order.
        for node in self.dli():
            yield node.key

    def values(self):

        # Return an iterator that returns the values in the cache in order
        # from the most recently to least recently used. Does not modify the
        # cache order.
        for node in self.dli():
            yield node.value

    def size(self, size=None):

        if size is not None:
            assert size > 0
            if size > self.listSize:
                self.addTailNode(size - self.listSize)
            elif size < self.listSize:
                self.removeTailNode(self.listSize - size)

        return self.listSize

    # Increases the size of the cache by inserting n empty nodes at the tail
    # of the list.
    def addTailNode(self, n):
        for i in range(n):
            node = _dlnode()
            node.next = self.head
            node.prev = self.head.prev

            self.head.prev.next = node
            self.head.prev = node

        self.listSize += n

    # Decreases the size of the list by removing n nodes from the tail of the
    # list.
    def removeTailNode(self, n):
        assert self.listSize > n
        for i in range(n):
            node = self.head.prev
            if not node.empty:
                if self.callback is not None:
                    self.callback(node.key, node.value)
                del self.table[node.key]

            # Splice the tail node out of the list
            self.head.prev = node.prev
            node.prev.next = self.head

            # The next four lines are not strictly necessary.
            node.prev = None
            node.next = None

            node.key = None
            node.value = None

        self.listSize -= n


    # This method adjusts the ordering of the doubly linked list so that
    # 'node' directly precedes the 'head' node. Because of the order of
    # operations, if 'node' already directly precedes the 'head' node or if
    # 'node' is the 'head' node the order of the list will be unchanged.
    def mtf(self, node):
        node.prev.next = node.next
        node.next.prev = node.prev

        node.prev = self.head.prev
        node.next = self.head.prev.next

        node.next.prev = node
        node.prev.next = node

    # This method returns an iterator that iterates over the non-empty nodes
    # in the doubly linked list in order from the most recently to the least
    # recently used.
    def dli(self):
        node = self.head
        for i in range(len(self.table)):
            yield node
            node = node.next




class WriteThroughCacheManager(object):
    def __init__(self, store, size):
        self.store = store
        self.cache = lrucache(size)

    def __len__(self):
        return len(self.store)

    # Returns/sets the size of the managed cache.
    def size(self, size=None):
        return self.cache.size(size)

    def clear(self):
        self.cache.clear()
        self.store.clear()

    def __contains__(self, key):
        # Check the cache first. If it is there we can return quickly.
        if key in self.cache:
            return True

        # Not in the cache. Might be in the underlying store.
        if key in self.store:
            return True

        return False

    def __getitem__(self, key):
        # First we try the cache. If successful we just return the value. If
        # not we catch KeyError and ignore it since that just means the key
        # was not in the cache.
        try:
            return self.cache[key]
        except KeyError:
            pass

        # It wasn't in the cache. Look it up in the store, add the entry to
        # the cache, and return the value.
        value = self.store[key]
        self.cache[key] = value
        return value

    def get(self, key, default=None):
        """Get an item - return default (None) if not present"""
        try:
            return self[key]
        except KeyError:
            return default

    def __setitem__(self, key, value):
        # Add the key/value pair to the cache and store.
        self.cache[key] = value
        self.store[key] = value

    def __delitem__(self, key):
        # Write-through behavior cache and store should be consistent. Delete
        # it from the store.
        del self.store[key]
        try:
            # Ok, delete from the store was successful. It might also be in
            # the cache, try and delete it. If not we catch the KeyError and
            # ignore it.
            del self.cache[key]
        except KeyError:
            pass

    def __iter__(self):
        return self.keys()

    def keys(self):
        return self.store.keys()

    def values(self):
        return self.store.values()

    def items(self):
        return self.store.items()



class WriteBackCacheManager(object):
    def __init__(self, store, size):
        self.store = store

        # Create a set to hold the dirty keys.
        self.dirty = set()

        # Define a callback function to be called by the cache when a
        # key/value pair is about to be ejected. This callback will check to
        # see if the key is in the dirty set. If so, then it will update the
        # store object and remove the key from the dirty set.
        def callback(key, value):
            if key in self.dirty:
                self.store[key] = value
                self.dirty.remove(key)

        # Create a cache and give it the callback function.
        self.cache = lrucache(size, callback)

    # Returns/sets the size of the managed cache.
    def size(self, size=None):
        return self.cache.size(size)

    def clear(self):
        self.cache.clear()
        self.dirty.clear()
        self.store.clear()

    def __contains__(self, key):
        # Check the cache first, since if it is there we can return quickly.
        if key in self.cache:
            return True

        # Not in the cache. Might be in the underlying store.
        if key in self.store:
            return True

        return False

    def __getitem__(self, key):
        # First we try the cache. If successful we just return the value. If
        # not we catch KeyError and ignore it since that just means the key
        # was not in the cache.
        try:
            return self.cache[key]
        except KeyError:
            pass

        # It wasn't in the cache. Look it up in the store, add the entry to
        # the cache, and return the value.
        value = self.store[key]
        self.cache[key] = value
        return value

    def get(self, key, default=None):
        """Get an item - return default (None) if not present"""
        try:
            return self[key]
        except KeyError:
            return default

    def __setitem__(self, key, value):
        # Add the key/value pair to the cache.
        self.cache[key] = value
        self.dirty.add(key)

    def __delitem__(self, key):

        found = False
        try:
            del self.cache[key]
            found = True
            self.dirty.remove(key)
        except KeyError:
            pass

        try:
            del self.store[key]
            found = True
        except KeyError:
            pass

        if not found:  # If not found in cache or store, raise error.
            raise KeyError


    def __iter__(self):
        return self.keys()

    def keys(self):
        for key in self.store.keys():
            if key not in self.dirty:
                yield key

        for key in self.dirty:
            yield key


    def values(self):
        for key, value in self.items():
            yield value


    def items(self):
        for key, value in self.store.items():
            if key not in self.dirty:
                yield (key, value)

        for key in self.dirty:
            value = self.cache.peek(key)
            yield (key, value)



    def sync(self):
        # For each dirty key, peek at its value in the cache and update the
        # store. Doesn't change the cache's order.
        for key in self.dirty:
            self.store[key] = self.cache.peek(key)
        # There are no dirty keys now.
        self.dirty.clear()

    def flush(self):
        self.sync()
        self.cache.clear()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.sync()
        return False





def lruwrap(store, size, writeback=False):
    if writeback:
        return WriteBackCacheManager(store, size)
    else:
        return WriteThroughCacheManager(store, size)




class lrudecorator(object):
    def __init__(self, size):
        self.cache = lrucache(size)

    def __call__(self, func):
        def wrapped(*args, **kwargs):
            kwtuple = tuple((key, kwargs[key]) for key in sorted(kwargs.keys()))
            key = (args, kwtuple)
            try:
                return self.cache[key]
            except KeyError:
                pass

            value = func(*args, **kwargs)
            self.cache[key] = value
            return value
        return wrapped



from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys
import codecs
import os
import re

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

class Tox(TestCommand):
    user_options = [('tox-args=', 'a', "Arguments to pass to tox")]
    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = None
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        import tox
        import shlex
        errno = tox.cmdline(args=shlex.split(self.tox_args))
        sys.exit(errno)

setup(
    name='distob',
    version=find_version('distob', '__init__.py'),
    url='http://github.com/mattja/distob/',
    license='GPLv3+',
    author='Matthew J. Aburn',
    install_requires=['IPython>=2.1',
                      'pyzmq>=2.1.11',
                      'dill>=0.2.1'],
    tests_require=['tox'],
    cmdclass = {'test': Tox},
    author_email='mattja6@gmail.com',
    description='Distributed computing made easier, using remote objects',
    long_description=read('README.rst'),
    packages=['distob'],
    platforms='any',
    zip_safe=False,
    classifiers = [
        'Programming Language :: Python',
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Distributed Computing',
        ],
    extras_require={'remote_arrays': ['numpy>=1.6']}
)



"""Remote and distributed numpy arrays
"""

from __future__ import absolute_import
from .distob import (proxy_methods, Remote, Ref, Error, 
                     scatter, gather, vectorize)
import numpy as np
from collections import Sequence
import numbers
import types
import warnings
import copy

# types for compatibility across python 2 and 3
_SliceType = type(slice(None))
_EllipsisType = type(Ellipsis)
_TupleType = type(())
_NewaxisType = type(np.newaxis)


def _brief_warning(msg, stacklevel=None):
    old_format = warnings.formatwarning
    def brief_format(message, category, filename, lineno, line=None):
        return '%s [%s:%d]\n' % (msg, filename, lineno)
    warnings.formatwarning = brief_format
    warnings.warn(msg, RuntimeWarning, stacklevel + 1)
    warnings.formatwarning = old_format


@proxy_methods(np.ndarray, exclude=('dtype',), include_underscore=(
    '__getitem__', '__setitem__', '__getslice__', '__setslice__'))
class RemoteArray(Remote, object):
    """Local object representing a remote ndarray"""
    def __init__(self, obj):
        """Make a RemoteArray to access an existing numpy.ndarray.

        Args:
          obj (Ref or object): either a Ref reference to the (possibly remote) 
            object to be controlled, or else an actual (local) object to be 
            controlled.
        """
        super(RemoteArray, self).__init__(obj)
        descr, shape, strides, typestr = self._ref.metadata
        # implement the python array interface
        self._array_intf = {'descr': descr, 'shape': shape, 'strides': strides,
                            'typestr': typestr, 'version': 3}
        self.dtype = np.dtype(typestr)

    #If a local consumer wants direct data access via the python 
    #array interface, then ensure a local copy of the data is in memory
    def __get_array_intf(self):
        if not self._obcache_current:
            msg = (u"Note: local numpy function requested local access to " +
                    "data. Fetching data..")
            _brief_warning(msg, stacklevel=4)
        self._fetch()
        self._array_intf['data'] = self._obcache.__array_interface__['data']
        return self._array_intf

    # This class cannot be a subclass of ndarray because that implies that
    # the C array interface is implemented. But the C interface promises
    # consumers they can just directly read memory of the array at any time.
    # Meaning array data is forced always to be in the local computer's memory.
    # Instead we implement only the python __array_interface__. Result is
    # all existing C code that calls PyArray_FromAny() on its inputs will
    # automatically work fine with these objects as if they were ndarrays.
    __array_interface__ = property(fget=__get_array_intf)

    shape = property(fget=lambda self: self._array_intf['shape'])

    strides = property(fget=lambda self: self._array_intf['strides'])

    ndim = property(fget=lambda self: len(self._array_intf['shape']))

    size = property(fget=lambda self: np.prod(self._array_intf['shape']))

    @classmethod
    def __pmetadata__(cls, obj):
        # obj is the real ndarray instance that we will control
        metadata = (obj.__array_interface__['descr'], obj.shape, 
                    obj.strides, obj.__array_interface__['typestr'])
        return metadata

    def __repr__(self):
        return self._cached_apply('__repr__').replace(
            'array', self.__class__.__name__, 1)

    def __len__(self):
        return self._array_intf['shape'][0]

    def __get_nbytes(self):
        if self._obcache is not None:
            return self._obcache.nbytes
        else:
            return 0

    nbytes = property(fget=__get_nbytes)

    def __array_finalize__(self, obj):
        #print('In proxy __array_finalize__, obj is type ' + str(type(obj)))
        if obj is None:
            return

    def __numpy_ufunc__(self, ufunc, method, i, inputs, **kwargs):
        """Route ufunc execution intelligently to local host or remote engine 
        depending on where the arguments reside.
        """
        print('In RemoteArray __numpy_ufunc__')
        print('ufunc:%s; method:%s; selfpos:%d' % (repr(ufunc), method, i))
        print('inputs:%s; kwargs:%s' % (inputs, kwargs))

        raise Error("ufunc=%s Haven't implemented ufunc support yet!" % ufunc)
        # TODO implement this!
        #return getattr(ufunc, method)(*inputs, **kwargs)

    def __array_prepare__(self, in_arr, context=None):
        """Fetch underlying data to user's computer and apply ufunc locally.
        Only used as a fallback, for numpy versions < 1.9 which lack 
        support for the __numpy_ufunc__ mechanism. 
        """
        print('In proxy __array_prepare__')
        #TODO fetch data here
        if map(int, np.version.short_version.split('.')) < [1,9,0]:
            raise Error('Numpy version 1.9.0 or later is required!')
        else:
            raise Error('Numpy is current, but still called __array_prepare__')
        #return super(RemoteArray, self).__array_prepare__(in_arr, context)

    def __array_wrap__(self, out_arr, context=None):
        print('In __array_wrap__')
        return np.ndarray.__array_wrap__(out_arr, context)

    def expand_dims(self, axis):
        """Insert a new axis, at a given position in the array shape
        Args:
          axis (int): Position (amongst axes) where new axis is to be inserted.
        """
        ix = [slice(None)] * self.ndim
        ix.insert(axis, np.newaxis)
        ix = tuple(ix)
        return self[ix]

    def concatenate(self, tup, axis=0):
        """Join a sequence of arrays to this one.

        Will aim to join `ndarray`, `RemoteArray`, and `DistArray` without
        moving their data if they are on different engines.

        Args: 
          tup (sequence of array_like): Arrays to be joined with this one.
            They must have the same shape as this array, except in the
            dimension corresponding to `axis`.
          axis (int, optional): The axis along which arrays will be joined.

        Returns:
          res `RemoteArray`, if arrays were all on the same engine
              `DistArray`, if inputs were on different engines
        """
        if not isinstance(tup, Sequence):
            tup = (tup,)
        if tup is (None,) or len(tup) is 0:
            return self
        tup = (self,) + tuple(tup)
        # convert all arguments to RemoteArrays if they are not already:
        arrays = []
        for ar in tup:
            if isinstance(ar, DistArray):
                if axis == ar._distaxis:
                    arrays.extend(ar._subarrays)
                else:
                    # Since not yet implemented arrays distributed on more than
                    # one axis, will fetch and re-scatter on the new axis.
                    ar = gather(ar)
                    arrays.extend(np.split(ar, ar.shape[axis], axis))
            elif isinstance(ar, RemoteArray):
                arrays.append(ar)
            elif isinstance(ar, Remote):
                arrays.append(_remote_to_array(ar))
            elif (not isinstance(ar, Remote) and 
                    not isinstance(ar, DistArray) and
                    not hasattr(type(ar), '__array_interface__')):
                arrays.append(np.array(ar))
        rarrays = scatter(arrays) # TODO: this will only work on the client
        # At this point rarrays is a list of RemoteArray to be concatenated
        refs = [ra._ref for ra in rarrays]
        eng_id = refs[0].engine_id
        if all(ref.engine_id == eng_id for ref in refs):
            # Arrays to be joined are all on the same engine
            from distob import engine
            if eng_id == engine.id:
                # All are on local host. Concatenate the local objects:
                return concatenate([engine[r.object_id] for r in refs], axis)
            else:
                # All are on the same remote host. Concatenate to a RemoteArray
                ids = [ref.object_id for ref in refs]
                dv = engine._client[eng_id]
                def remote_concat(ids, axis):
                    from distob import engine
                    ar = np.concatenate([engine[id] for id in ids], axis)
                    return Ref(ar)
                ref = dv.apply_sync(remote_concat, ids, axis)
                return RemoteArray(ref)
        else:
            # Arrays to be joined are on different engines.
            concatenation_axis_lengths = [ra.shape[axis] for ra in rarrays]
            if not all(n == 1 for n in concatenation_axis_lengths):
                #TODO: should do it remotely, using distob.split()
                msg = (u'Concatenating remote axes with lengths other than 1.'
                        'current implementation will fetch all data locally!')
                warnings.warn(msg, RuntimeWarning)
                array = concatenate(gather(rarrays), axis)
                return scatter(array, axis)
            else:
                return DistArray(rarrays, axis)


class DistArray(object):
    """Local object representing a single array distributed on multiple engines

    Currently only one of the axes will be distributed.
    """
    def __init__(self, subarrays, axis=None):
        """Make a DistArray from a list of existing remote numpy.ndarrays

        Args:
          subarrays (list of RemoteArray, or list of Ref to ndarray): 
            the subarrays (possibly remote) which form the whole array when
            concatenated. The subarrays must all have the same shape and dtype.
            Currently must have `a.shape[axis] == 1` for each subarray `a`
          axis (int, optional): Position of the distributed axis, which is the 
            axis along which the subarrays will be concatenated. Default is 
            the last axis.  
        """
        self._n = len(subarrays)  # Length of the distributed axis.
        if self._n < 2:
            raise ValueError('must provide more than one subarray')
        self._obcache = None
        self._obcache_current = False
        # In the subarrays list, accept either RemoteArray or Ref to ndarray:
        subarrays = [ra if isinstance(ra, RemoteArray)
                     else RemoteArray(ra) for ra in subarrays]
        first_ra_metadata = subarrays[0]._ref.metadata
        if not all(ra._ref.metadata == first_ra_metadata for ra in subarrays):
            raise ValueError('subarrays must have same shape, strides & dtype')
        descr, subshape, substrides, typestr = first_ra_metadata
        itemsize = int(typestr[2:])
        if axis is None:
            self._distaxis = len(subshape) - 1
        else:
            self._distaxis = axis
        # For numpy strides, `None` means assume C-style ordering:
        if substrides is None:
            substrides = tuple(int(np.product(
                    subshape[i:])*itemsize) for i in range(1, len(subshape)+1))
        shape = list(subshape)
        shape[self._distaxis] = self._n
        shape = tuple(shape)
        # For now, report the same strides as used by the subarrays
        strides = substrides
        self.dtype = np.dtype(typestr)
        self._subarrays = subarrays
        # a surrogate ndarray to help with slicing of the distributed axis:
        self._placeholders = np.array(range(len(subarrays)), dtype=int)
        # implement the python array interface
        self._array_intf = {'descr': descr, 'shape': shape, 'strides': strides,
                            'typestr': typestr, 'version': 3}

    def _fetch(self):
        """update local cached copy of the real object"""
        if not self._obcache_current:
            ax = self._distaxis
            self._obcache = concatenate([ra._ob for ra in self._subarrays], ax)
            self._obcache_current = True

    def __ob(self):
        """return a copy of the real object"""
        self._fetch()
        return self._obcache

    _ob = property(fget=__ob)

    #If a local consumer asks for direct data access via the python 
    #array interface, attempt to put a local copy of all the data into memory
    def __get_array_intf(self):
        #print('__array_interface__ requested.')
        self._fetch()
        self._array_intf['data'] = self._obcache.__array_interface__['data']
        self._array_intf['strides'] = \
                self._obcache.__array_interface__['strides']
        return self._array_intf

    __array_interface__ = property(fget=__get_array_intf)

    shape = property(fget=lambda self: self._array_intf['shape'])

    size = property(fget=lambda self: np.prod(self._array_intf['shape']))

    strides = property(fget=lambda self: self._array_intf['strides'])

    def __repr__(self):
        classname = self.__class__.__name__
        selectors = []
        for i in range(self.ndim):
            if self.shape[i] > 3:
                selectors.append((0, 1, -2, -1))
            else:
                selectors.append(tuple(range(self.shape[i])))
        ix = tuple(np.meshgrid(*selectors, indexing='ij'))
        corners = gather(self[ix])
        def _repr_nd(corners, shape, indent):
            """Recursively generate abbreviated text representation of array"""
            if not shape:
                if isinstance(corners, np.float):
                    return '{: .8f}'.format(corners)
                else:
                    return str(corners)
            else:
                if len(shape) > 1:
                    pre = ',' + '\n'*(len(shape)-1) + ' '*indent
                else:
                    pre = ', '
                s = '['
                n = shape[0]
                if n > 0:
                    s += _repr_nd(corners[0], shape[1:], indent+1)
                if n > 1:
                    s += pre + _repr_nd(corners[1], shape[1:], indent+1)
                if n > 4:
                    s += pre + '...'
                if n > 3:
                    s += pre + _repr_nd(corners[-2], shape[1:], indent+1)
                if n > 2:
                    s += pre + _repr_nd(corners[-1], shape[1:], indent+1)
                s += ']'
                return s
        s = u'<%s of shape %s with axis %d distributed>:\n' % (
                classname, self.shape, self._distaxis)
        indent = len(classname) + len(self.shape) - 1
        s += classname + '(' + _repr_nd(corners, self.shape, indent) + ')'
        return s

    def __len__(self):
        return self._array_intf['shape'][0]

    def __get_nbytes(self):
        return sum(ra.nbytes for ra in self._subarrays)

    nbytes = property(fget=__get_nbytes)

    ndim = property(fget=lambda self: len(self._array_intf['shape']))

    def __array_finalize__(self, obj):
        #print('In DistArray __array_finalize__, obj type: ' + str(type(obj)))
        if obj is None:
            return

    def __numpy_ufunc__(self, ufunc, method, i, inputs, **kwargs):
        """Route ufunc execution intelligently to local host or remote engine 
        depending on where the arguments reside.
        """
        print('In DistArray __numpy_ufunc__')
        print('ufunc:%s; method:%s; selfpos:%d' % (repr(ufunc), method, i))
        print('inputs:%s; kwargs:%s' % (inputs, kwargs))
        raise Error("ufunc=%s Haven't implemented ufunc support yet!" % ufunc)

    def __array_prepare__(self, in_arr, context=None):
        """Fetch underlying data to user's computer and apply ufunc locally.
        Only used as a fallback, for numpy versions < 1.9 which lack 
        support for the __numpy_ufunc__ mechanism. 
        """
        print('In DistArray __array_prepare__')
        #TODO fetch data here
        if map(int, np.version.short_version.split('.')) < [1,9,0]:
            raise Error('Numpy version 1.9.0 or later is required.')
        else:
            raise Error('Numpy is current, but still called __array_prepare__')

    def __array_wrap__(self, out_arr, context=None):
        print('In DistArray __array_wrap__')
        return np.ndarray.__array_wrap__(out_arr, context)

    def __getitem__(self, index):
        """Slice the distributed array"""
        # To be a DistArray, must have an axis across >=2 engines. If the slice
        # result will no longer be distributed, we return a RemoteArray instead
        distaxis = self._distaxis
        if isinstance(index, np.ndarray) and index.dtype.type is np.bool_:
            raise Error('indexing by boolean array not yet implemented')
        if not isinstance(index, Sequence):
            index = (index,) + (slice(None),)*(self.ndim - 1)
        ix_types = tuple(type(x) for x in index)
        if (np.ndarray in ix_types or
                (not isinstance(index, _TupleType) and 
                    _NewaxisType not in ix_types and 
                    _EllipsisType not in ix_types and
                    _SliceType not in ix_types) or
                any(issubclass(T, Sequence) for T in ix_types)):
            basic_slicing = False
        else:
            basic_slicing = True
        # Apply any ellipses
        while _EllipsisType in ix_types:
            pos = ix_types.index(_EllipsisType)
            m = (self.ndim + ix_types.count(_NewaxisType) - len(index) + 1)
            index = index[:pos] + (slice(None),)*m + index[(pos+1):]
            ix_types = tuple(type(x) for x in index)
        # Apply any np.newaxis
        if _NewaxisType in ix_types:
            new_distaxis = distaxis
            subix = [slice(None)] * self.ndim
            while _NewaxisType in ix_types:
                pos = ix_types.index(type(np.newaxis))
                index = index[:pos] + (slice(None),) + index[(pos+1):]
                ix_types = tuple(type(x) for x in index)
                if pos <= distaxis:
                    subix[pos] = np.newaxis
                    new_distaxis += 1
                else:
                    subix[pos - 1] = np.newaxis
            new_subarrays = [ra[tuple(subix)] for ra in self._subarrays]
            return DistArray(new_subarrays, new_distaxis)[index]
        index = tuple(index) + (slice(None),)*(self.ndim - len(index))
        if len(index) > self.ndim:
            raise IndexError('too many indices for array')
        if basic_slicing:
            # separate the index acting on distributed axis from other indices
            distix = index[distaxis]
            ixlist = self._placeholders[distix]
            if isinstance(ixlist, numbers.Number):
                # distributed axis has been sliced away: return a RemoteArray
                subix = index[0:distaxis] + (0,) + index[(distaxis+1):]
                return self._subarrays[ixlist][subix]
            some_subarrays = [self._subarrays[i] for i in ixlist]
            subix = index[0:distaxis] + (slice(None),) + index[(distaxis+1):]
            result_ras = [ra[subix] for ra in some_subarrays]
            if len(result_ras) is 1:
                # no longer distributed: return a RemoteArray
                return result_ras[0]
            else:
                axes_removed = sum(1 for x in index[:distaxis] if isinstance(
                        x, numbers.Integral))
                new_distaxis = distaxis - axes_removed
                return DistArray(result_ras, new_distaxis)
        else:
            # advanced integer slicing
            is_fancy = tuple(not isinstance(x, _SliceType) for x in index)
            fancy_pos = tuple(i for i in range(len(index)) if is_fancy[i])
            slice_pos = tuple(i for i in range(len(index)) if not is_fancy[i])
            contiguous = (fancy_pos[-1] - fancy_pos[0] == len(fancy_pos) - 1)
            index = list(index)
            ix_arrays = [index[j] for j in fancy_pos]
            ix_arrays = np.broadcast_arrays(*ix_arrays)
            for j in range(len(fancy_pos)):
                index[fancy_pos[j]] = ix_arrays[j]
            index = tuple(index)
            idim = index[fancy_pos[0]].ndim # common ndim of all index arrays
            assert(idim > 0)
            distix = index[distaxis]
            otherix = index[0:distaxis] + (slice(None),) + index[(distaxis+1):]
            if not is_fancy[distaxis]:
                # fancy indexing is only being applied to non-distributed axes
                ixlist = self._placeholders[distix]
                if isinstance(ixlist, numbers.Number):
                    # distributed axis was sliced away: return a RemoteArray
                    subix = index[0:distaxis] + (0,) + index[(distaxis+1):]
                    return self._subarrays[ixlist][subix]
                some_subarrays = [self._subarrays[i] for i in ixlist]
                # predict where that new axis will be in subarrays post-slicing
                if contiguous:
                    if fancy_pos[0] > distaxis:
                        new_distaxis = distaxis
                    else:
                        new_distaxis = distaxis - len(fancy_pos) + idim
                else:
                    earlier_fancy = len([i for i in fancy_pos if i < distaxis])
                    new_distaxis = distaxis - earlier_fancy + idim
                result_ras = [ra[otherix] for ra in some_subarrays]
            else:
                # fancy indexing is being applied to the distributed axis
                nonconstant_ix_axes = []
                for j in range(idim):
                    n = distix.shape[j]
                    if n > 1:
                        partix = np.split(distix, n, axis=j)
                        if not all(np.array_equal(
                                partix[0], partix[i]) for i in range(1, n)):
                            nonconstant_ix_axes.append(j)
                if len(nonconstant_ix_axes) <= 1:
                    # then we can apply the indexing without moving data
                    if len(nonconstant_ix_axes) is 0:
                        # implies every entry of indexing array for distaxis is
                        # the same, so all result data is on a single engine.
                        all_same_engine = True
                        iax = idim - 1
                    else:
                        # len(nonconstant_ix_axes) is 1
                        # the result will remain distributed on one output axis
                        all_same_engine = False
                        iax = nonconstant_ix_axes[0]
                    iix = [0] * idim
                    iix[iax] = slice(None)
                    iix = tuple(iix)
                    ixlist = self._placeholders[distix[iix]]
                    some_subarrays = [self._subarrays[i] for i in ixlist]
                    if contiguous:
                        new_distaxis = fancy_pos[0] + iax
                    else:
                        new_distaxis = 0 + iax
                    result_ras = []
                    for i in range(distix.shape[iax]):
                        # Slice the original indexing arrays into smaller
                        # arrays, suitable for indexing our subarrays.
                        sl = [slice(None)] * idim
                        sl[iax] = slice(i, i+1)
                        sl = tuple(sl)
                        subix = list(index)
                        for j in range(len(subix)):
                            if isinstance(subix[j], np.ndarray):
                                subix[j] = subix[j][sl]
                                if j == distaxis:
                                    subix[j] = np.zeros_like(subix[j])
                        subix = tuple(subix)
                        result_ras.append(some_subarrays[i][subix])
                    if all_same_engine and len(result_ras) > 1:
                        rs = [expand_dims(r, new_distaxis) for r in result_ras]
                        return concatenate(rs, new_distaxis) # one RemoteArray
                else:
                    # Have more than one nonconstant indexing axis in the
                    # indexing array that applies to our distributed axis.
                    msg = (u'The requested slicing operation requires moving '
                            'data between engines. Will fetch data locally.')
                    warnings.warn(msg, RuntimeWarning)
                    return self._ob[index]
            if len(result_ras) is 1:
                # no longer distributed: return a RemoteArray
                return result_ras[0]
            else:
                return DistArray(result_ras, new_distaxis)

    def __setitem__(self, index, value):
        """Assign to the sliced item"""
        raise Error(u'assigning to remote arrays via slices is '
                     'not yet implemented. stay tuned.')

    def transpose(self, *axes):
        """Returns a view of the array with axes transposed.

        For a 1-D array, this has no effect.
        For a 2-D array, this is the usual matrix transpose.
        For an n-D array, if axes are given, their order indicates how the
        axes are permuted

        Args:
          a (array_like): Input array.
          axes (tuple of int, optional): By default, reverse the dimensions, 
            otherwise permute the axes according to values given.
        """
        if axes is ():
            axes = None
        return transpose(self, axes)

    T = property(fget=transpose)

    @classmethod
    def __distob_vectorize__(cls, f):
        """Upgrades a normal function f to act on a DistArray in parallel

        Args:
          f (callable): ordinary function which expects as its first 
            argument an array (of the same shape as our subarrays)

        Returns:
          vf (callable): new function that takes a DistArray as its first
            argument. ``vf(distarray)`` will do the computation ``f(subarray)``
            on each subarray in parallel and if possible will return another 
            DistArray. (otherwise will return a list with the result for each 
            subarray).
        """
        def vf(self, *args, **kwargs):
            remove_axis = ((slice(None),)*(self._distaxis) + (0,) + 
                           (slice(None),)*(self.ndim - self._distaxis - 1))
            refs = [ra[remove_axis]._ref for ra in self._subarrays]
            from distob import engine
            dv = engine._client[:]
            def remote_f(object_id, *args, **kwargs):
                import distob
                result = f(distob.engine[object_id], *args, **kwargs)
                if type(result) in distob.engine.proxy_types:
                    return Ref(result)
                else:
                    return result
            results = []
            for ref in refs:
                dv.targets = ref.engine_id
                ar = dv.apply_async(remote_f, ref.object_id, *args, **kwargs)
                results.append(ar)
            for i in range(len(results)):
                ar = results[i]
                ar.wait()
                results[i] = ar.r
                if isinstance(results[i], Ref):
                    ref = results[i]
                    RemoteClass = engine.proxy_types[ref.type]
                    results[i] = RemoteClass(ref)
            if (all(isinstance(r, RemoteArray) for r in results) and
                    all(r.shape == results[0].shape for r in results)):
                # Then we can join the results and return a DistArray.
                # To position result distaxis, match input shape where possible
                old_subshape = (self.shape[0:self._distaxis] +
                                self.shape[(self._distaxis+1):])
                res_subshape = list(results[0].shape)
                pos = len(res_subshape)
                for i in range(len(old_subshape)):
                    n = old_subshape[i]
                    if n not in res_subshape:
                        continue
                    pos = res_subshape.index(n)
                    res_subshape[pos] = None
                    if i >= self._distaxis:
                        break
                    pos += 1
                new_distaxis = pos
                results = [r.expand_dims(new_distaxis) for r in results]
                return DistArray(results, new_distaxis)
            elif all(isinstance(r, numbers.Number) for r in results):
                return np.array(results)
            else:
                return results  # list
        if hasattr(f, '__name__'):
            vf.__name__ = 'v' + f.__name__
            f_str = f.__name__ + '()'
        else:
            f_str = 'callable'
        doc = u"""Apply %s in parallel to a DistArray\n
               Args:
                 da (DistArray)
                 other args are the same as for %s
               """ % (f_str, f_str)
        if hasattr(f, '__doc__') and f.__doc__ is not None:
            doc = doc.rstrip() + (' detailed below:\n----------\n' + f.__doc__)
        vf.__doc__ = doc
        return vf

    def __distob_gather__(self):
        return self._ob

    def expand_dims(self, axis):
        """Insert a new axis, at a given position in the array shape
        Args:
          axis (int): Position (amongst axes) where new axis is to be inserted.
        """
        if axis <= self._distaxis:
            subaxis = axis
            new_distaxis = self._distaxis + 1
        else:
            subaxis = axis - 1
            new_distaxis = self._distaxis
        new_subarrays = [expand_dims(ra, subaxis) for ra in self._subarrays]
        return DistArray(new_subarrays, new_distaxis)

    def concatenate(self, tup, axis=0):
        """Join a sequence of arrays to this one.

        Will aim to join `ndarray`, `RemoteArray`, and `DistArray` without
        moving their data if they are on different engines.

        Args: 
          tup (sequence of array_like): Arrays to be joined with this one.
            They must have the same shape as this array, except in the
            dimension corresponding to `axis`.
          axis (int, optional): The axis along which arrays will be joined.

        Returns:
          res `RemoteArray`, if arrays were all on the same engine
              `DistArray`, if inputs were on different engines
        """
        if not isinstance(tup, Sequence):
            tup = (tup,)
        if tup is (None,) or len(tup) is 0:
            return self
        if axis == self._distaxis:
            split_self = self._subarrays
        else:
            # Since we have not yet implemented arrays distributed on more than
            # one axis, will fetch subarrays and re-scatter on the new axis.
            split_self = split(self._ob, self.shape[axis], axis)
        first = split_self[0]
        others = tuple(split_self[1:]) + tup
        first = scatter(first)
        return first.concatenate(others, axis)

    def mean(self, axis=None, dtype=None, out=None):
        """Compute the arithmetic mean along the specified axis."""
        if axis is None:
            if out is not None:
                out[:] = vectorize(np.mean)(self, None, dtype).mean()
                return out
            else:
                return vectorize(np.mean)(self, None, dtype).mean()
        elif axis == self._distaxis:
            return np.mean(self._ob, axis, dtype, out)
        else:
            if axis > self._distaxis:
                axis -= 1
            if out is not None:
                out[:] = vectorize(np.mean)(self, axis, dtype)
                return out
            else:
                return vectorize(np.mean)(self, axis, dtype)


def transpose(a, axes=None):
    """Returns a view of the array with axes transposed.

    For a 1-D array, this has no effect.
    For a 2-D array, this is the usual matrix transpose.
    For an n-D array, if axes are given, their order indicates how the
    axes are permuted

    Args:
      a (array_like): Input array.
      axes (list of int, optional): By default, reverse the dimensions,
        otherwise permute the axes according to the values given.
    """
    if isinstance(a, np.ndarray):
        return np.transpose(a, axes)
    elif isinstance(a, RemoteArray):
        return a.transpose(*axes)
    elif isinstance(a, Remote):
        return _remote_to_array(a).transpose(*axes)
    elif isinstance(a, DistArray):
        if axes is None:
            axes = range(a.ndim - 1, -1, -1)
        axes = list(axes)
        if len(set(axes)) < len(axes):
            raise ValueError("repeated axis in transpose")
        if sorted(axes) != list(range(a.ndim)):
            raise ValueError("axes don't match array")
        distaxis = a._distaxis
        new_distaxis = axes.index(distaxis)
        new_subarrays = [ra.transpose(*axes) for ra in a._subarrays]
        return DistArray(new_subarrays, new_distaxis)
    else:
        return np.transpose(a, axes)


def rollaxis(a, axis, start=0):
    """Roll the specified axis backwards, until it lies in a given position.

    Args:
      a (array_like): Input array.
      axis (int): The axis to roll backwards.  The positions of the other axes 
        do not change relative to one another.
      start (int, optional): The axis is rolled until it lies before this 
        position.  The default, 0, results in a "complete" roll.

    Returns:
      res (ndarray)
    """
    if isinstance(a, np.ndarray):
        return np.rollaxis(a, axis, start)
    if axis not in range(a.ndim):
        raise ValueError(
                'rollaxis: axis (%d) must be >=0 and < %d' % (axis, a.ndim))
    if start not in range(a.ndim + 1):
        raise ValueError(
                'rollaxis: start (%d) must be >=0 and < %d' % (axis, a.ndim+1))
    axes = list(range(a.ndim))
    axes.remove(axis)
    axes.insert(start, axis)
    return transpose(a, axes)


def expand_dims(a, axis):
    """Insert a new axis, corresponding to a given position in the array shape

    Args:
      a (array_like): Input array.
      axis (int): Position (amongst axes) where new axis is to be inserted.
    """
    if hasattr(a, 'expand_dims') and hasattr(type(a), '__array_interface__'):
        return a.expand_dims(axis)
    if isinstance(a, np.ndarray):
        return np.expand_dims(a, axis)


def _remote_to_array(remote):
    # Try to convert an arbitrary Remote object into a RemoteArray
    if isinstance(remote, RemoteArray):
        return remote
    else:
        from distob import engine
        dv = engine._client[remote._ref.engine_id]
        def remote_array(object_id):
            return Ref(np.array(engine[object_id]))
        return RemoteArray(dv.apply_sync(remote_array, remote._ref.object_id))


def concatenate(tup, axis=0):
    """Join a sequence of arrays together. 
    Will aim to join `ndarray`, `RemoteArray`, and `DistArray` without moving 
    their data, if they happen to be on different engines.

    Args:
      tup (sequence of array_like): Arrays to be concatenated. They must have
        the same shape, except in the dimension corresponding to `axis`.
      axis (int, optional): The axis along which the arrays will be joined.

    Returns: 
      res: `ndarray`, if inputs were all local
           `RemoteArray`, if inputs were all on the same remote engine
           `DistArray`, if inputs were already scattered on different engines
    """
    if len(tup) is 0:
        raise ValueError('need at least one array to concatenate')
    first = tup[0]
    others = tup[1:]
    if (hasattr(first, 'concatenate') and 
            hasattr(type(first), '__array_interface__')):
        # This case includes RemoteArray and DistArray
        return first.concatenate(others, axis)
    if all(isinstance(ar, np.ndarray) for ar in tup):
        return np.concatenate(tup, axis)
    if isinstance(first, Remote):
        first = _remote_to_array(first)
    else:
        first = scatter(np.array(first)) #TODO: this only works on the client
    return first.concatenate(others, axis)


def vstack(tup):
    """Stack arrays in sequence vertically (row wise), 
    handling ``RemoteArray`` and ``DistArray`` without moving data.

    Args:
      tup (sequence of array_like)

    Returns: 
      res: `ndarray`, if inputs were all local
           `RemoteArray`, if inputs were all on the same remote engine
           `DistArray`, if inputs were already scattered on different engines
    """
    # Follow numpy.vstack behavior for 1D arrays:
    arrays = list(tup)
    for i in range(len(arrays)):
        if arrays[i].ndim is 1:
            arrays[i] = arrays[i][np.newaxis, :]
    return concatenate(tup, axis=0)


def hstack(tup):
    """Stack arrays in sequence horizontally (column wise), 
    handling ``RemoteArray`` and ``DistArray`` without moving data.

    Args:
      tup (sequence of array_like)

    Returns: 
      res: `ndarray`, if inputs were all local
           `RemoteArray`, if inputs were all on the same remote engine
           `DistArray`, if inputs were already scattered on different engines
    """
    # Follow numpy.hstack behavior for 1D arrays:
    if all(ar.ndim is 1 for ar in tup):
        return concatenate(tup, axis=0)
    else:
        return concatenate(tup, axis=1)


def dstack(tup):
    """Stack arrays in sequence depth wise (along third dimension), 
    handling ``RemoteArray`` and ``DistArray`` without moving data.

    Args:
      tup (sequence of array_like)

    Returns: 
      res: `ndarray`, if inputs were all local
           `RemoteArray`, if inputs were all on the same remote engine
           `DistArray`, if inputs were already scattered on different engines
    """
    # Follow numpy.dstack behavior for 1D and 2D arrays:
    arrays = list(tup)
    for i in range(len(arrays)):
        if arrays[i].ndim is 1:
            arrays[i] = arrays[i][np.newaxis, :]
        if arrays[i].ndim is 2:
            arrays[i] = arrays[i][:, :, np.newaxis]
    return concatenate(arrays, axis=2)


def split(a, indices_or_sections, axis=0):
    if type(a) is np.ndarray: # deliberately excluding subclasses
        return np.split(a, indices_or_sections, axis)
    else:
        if not isinstance(indices_or_sections, numbers.Integral):
            raise Error('splitting by array of indices is not yet implemented')
        n = indices_or_sections
        if self.shape[axis] % n != 0:
            raise ValueError("Array split doesn't result in an equal division")
        step = self.shape[axis] / n
        pieces = []
        start = 0
        while start < self.shape[axis]:
            stop = start + step
            ix = [slice(None)] * self.ndim
            ix[axis] = slice(start, stop)
            ix = tuple(ix)
            pieces.append(self[ix])
            start += step
        return pieces


def vsplit(a, indices_or_sections):
    return split(a, indices_or_sections, axis=0)


def hsplit(a, indices_or_sections):
    return split(a, indices_or_sections, axis=1)


def dsplit(a, indices_or_sections):
    return split(a, indices_or_sections, axis=2)


def broadcast_arrays(*args):
    if all(type(a) is np.ndarray for a in args): # exclude subclasses
        return np.broadcast_arrays(*args)
    shapes = [a.shape for a in args]
    ndim = max(len(sh) for sh in shapes) # new common ndim after broadcasting
    arrays = list(args)
    for i in range(len(arrays)):
        old_ndim = len(shapes[i])
        if old_ndim < ndim:
            ix = (np.newaxis,)*(ndim - old_ndim) + (slice(None),)*old_ndim
            arrays[i] = arrays[i][ix]
            shapes[i] = arrays[i].shape
    newshape = tuple(max(sh[ax] for sh in shapes) for ax in range(ndim))
    # Now broadcast arrays using fancy indexing. This is wasteful of both
    # memory and CPU, as it actually creates the redundant arrays instead of
    # making views with zero stride, but it works universally on classes that
    # implement fancy indexing. (TODO: Broadcast efficiently by having each
    # class implement its own `broaden(newshape)` method that returns views)
    for i in range(len(arrays)):
        for ax in range(ndim):
            if shapes[i][ax] == newshape[ax]:
                pass
            elif shapes[i][ax] == 1:
                ix = ((slice(None),) * ax + 
                      (np.zeros(newshape[ax], dtype=np.int64),) + 
                      (slice(None),) * (ndim - ax - 1))
                arrays[i] = arrays[i][ix]
            else:
                raise ValueError(u'shape mismatch: two or more arrays have '
                                  'incompatible dimensions on axis %d' % ax)
    return arrays
